# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_identification import IgnoreLocationException
from colcon_core.package_identification import logger
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.package_identification.python import get_configuration
from colcon_core.package_identification.python import is_reading_cfg_sufficient
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from colcon_python_setup_py.package_identification.python_setup_py \
    import get_setup_information


# mapping paths to tuples containing the ROS package and its build type
_cached_packages = {}


class RosPackageIdentification(
    PackageIdentificationExtensionPoint, PackageAugmentationExtensionPoint
):
    """Identify ROS packages with `package.xml` files."""

    # the priority needs to be higher than the extensions identifying packages
    # using the build systems supported by ROS (CMake and Python)
    PRIORITY = 150

    def __init__(self):  # noqa: D107
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        # check if the necessary dependency to parse the manifest is available
        try:
            import catkin_pkg  # noqa: F401
        except ImportError:
            raise SkipExtensionException(
                "The Python module 'catkin_pkg' could not be imported, "
                'therefore ROS packages can not be identified')

    def identify(self, desc):  # noqa: D102
        # ignore packages which have been identified with a different type
        if desc.type is not None and desc.type != 'ros':
            return

        # skip paths with an ignore marker file
        if (desc.path / 'CATKIN_IGNORE').exists():
            raise IgnoreLocationException()
        if (desc.path / 'AMENT_IGNORE').exists():
            raise IgnoreLocationException()

        # parse package manifest and get build type
        pkg, build_type = get_package_with_build_type(str(desc.path))
        if not pkg or not build_type:
            # if it is not a wet ROS package check for a dry ROS package
            if (desc.path / 'manifest.xml').exists():
                # ignore location to avoid being identified as a CMake package
                raise IgnoreLocationException()
            return

        # for Python build types ensure that a setup.py file exists
        if build_type == 'ament_python':
            setup_py = desc.path / 'setup.py'
            if not setup_py.is_file():
                logger.error(
                    "ROS package '{desc.path}' with build type '{build_type}' "
                    "has no 'setup.py' file" .format_map(locals()))
                raise IgnoreLocationException()

        desc.type = 'ros.{build_type}'.format_map(locals())

        # use package name from manifest if not already set
        # e.g. from external configuration
        if desc.name is None:
            desc.name = pkg.name

        desc.metadata['version'] = pkg.version

        # get dependencies
        for d in pkg.build_depends + pkg.buildtool_depends:
            assert d.evaluated_condition is not None
            if d.evaluated_condition:
                desc.dependencies['build'].add(DependencyDescriptor(
                    d.name, metadata=_create_metadata(d)))

        for d in (
            pkg.build_export_depends +
            pkg.buildtool_export_depends +
            pkg.exec_depends
        ):
            assert d.evaluated_condition is not None
            if d.evaluated_condition:
                desc.dependencies['run'].add(DependencyDescriptor(
                    d.name, metadata=_create_metadata(d)))

        for d in pkg.test_depends:
            assert d.evaluated_condition is not None
            if d.evaluated_condition:
                desc.dependencies['test'].add(DependencyDescriptor(
                    d.name, metadata=_create_metadata(d)))

        # for Python build types ensure that a setup.py file exists
        if build_type == 'ament_python':
            setup_cfg = desc.path / 'setup.cfg'
            for _ in (1, ):
                # try to get information from setup.cfg file
                if setup_cfg.is_file():
                    if is_reading_cfg_sufficient(setup_py):
                        config = get_configuration(setup_cfg)
                        name = config.get('metadata', {}).get('name')
                        if name:
                            options = config.get('options', {})

                            def getter(env):
                                nonlocal options
                                return options
                            break
            else:
                # use information from setup.py file

                def getter(env):  # noqa: F811
                    nonlocal desc
                    return get_setup_information(
                        desc.path / 'setup.py', env=env)

            desc.metadata['get_python_setup_options'] = getter

    def augment_packages(
        self, descs, *, additional_argument_names=None
    ):  # noqa: D102
        # get all parsed ROS package manifests
        global _cached_packages
        pkgs = {}
        for desc in descs:
            if str(desc.path) not in _cached_packages:
                continue
            pkg = _cached_packages[str(desc.path)][0]
            if pkg:
                pkgs[pkg] = desc

        # resolve group members and add them to the descriptor dependencies
        for pkg, desc in pkgs.items():
            for group_depend in pkg.group_depends:
                assert group_depend.evaluated_condition is not None
                if not group_depend.evaluated_condition:
                    continue
                group_depend.extract_group_members(pkgs)
                for name in group_depend.members:
                    desc.dependencies['build'].add(DependencyDescriptor(name))
                    desc.dependencies['run'].add(DependencyDescriptor(name))


def get_package_with_build_type(path: str):
    """Get the ROS package and its build type for the given path."""
    global _cached_packages
    if path not in _cached_packages:
        pkg = _get_package(path)
        build_type = _get_build_type(pkg) if pkg else None
        _cached_packages[path] = (pkg, build_type)
    return _cached_packages[path]


def _get_package(path: str):
    """Get the ROS package for the given path."""
    from catkin_pkg.package import has_ros_schema_reference
    from catkin_pkg.package import InvalidPackage
    from catkin_pkg.package import package_exists_at
    from catkin_pkg.package import parse_package

    if not package_exists_at(path):
        return None

    try:
        pkg = parse_package(path)
    except (AssertionError, InvalidPackage) as e:  # noqa: F841
        if has_ros_schema_reference(path):
            logger.debug(
                "Found ROS schema reference in package manifest in '{path}'"
                .format_map(locals()))
            logger.warn(
                "Failed to parse ROS package manifest in '{path}': {e}"
                .format_map(locals()))
        else:
            logger.debug(
                'Failed to parse potential ROS package manifest in'
                "'{path}': {e}".format_map(locals()))
        return None

    pkg.evaluate_conditions(os.environ)
    return pkg


def _get_build_type(pkg):
    """Get the build type of the ROS package."""
    from catkin_pkg.package import InvalidPackage
    try:
        return pkg.get_build_type()
    except InvalidPackage:
        logger.warning(
            "ROS package '{pkg.name}' in '{path}' has more than one "
            'build type'.format_map(locals()))
        return None


def _create_metadata(dependency):
    metadata = {}
    attributes = (
        'version_lte',
        'version_lt',
        'version_gte',
        'version_gt',
        'version_eq',
    )
    for attr in attributes:
        if getattr(dependency, attr, None) is not None:
            metadata[attr] = getattr(dependency, attr)
    return metadata



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import warnings


def add_app_to_cpp(env):
    """
    Add AMENT_PREFIX_PATH to CMAKE_PREFIX_PATH.

    Each ament prefix path is inserted before the first catkin prefix path or
    at the end if none of the CMake prefix paths has a '.catkin' marker file.
    """
    ament_prefix_path = os.environ.get('AMENT_PREFIX_PATH')
    if ament_prefix_path:
        cmake_prefix_path = env.get('CMAKE_PREFIX_PATH')
        cpp = cmake_prefix_path.split(os.pathsep) if cmake_prefix_path else []
        cpp_has_dot_catkin = [
            os.path.exists(os.path.join(p, '.catkin')) for p in cpp]
        app = ament_prefix_path.split(os.pathsep)
        for p in app:
            if p not in cpp:
                try:
                    index = cpp_has_dot_catkin.index(True)
                    cpp.insert(index, p)
                    cpp_has_dot_catkin.insert(index, False)
                except ValueError:
                    cpp.append(p)
        env['CMAKE_PREFIX_PATH'] = os.pathsep.join(cpp)


def append_app_to_cpp(env):
    """Append AMENT_PREFIX_PATH to CMAKE_PREFIX_PATH."""
    warnings.warn(
        'colcon_ros.task.append_app_to_cpp() will be removed in the future',
        DeprecationWarning, stacklevel=2)
    ament_prefix_path = os.environ.get('AMENT_PREFIX_PATH')
    if ament_prefix_path:
        cmake_prefix_path = env.get('CMAKE_PREFIX_PATH')
        cpp = cmake_prefix_path.split(os.pathsep) if cmake_prefix_path else []
        app = ament_prefix_path.split(os.pathsep)
        for p in app:
            if p not in cpp:
                cpp.append(p)
        env['CMAKE_PREFIX_PATH'] = os.pathsep.join(cpp)


def extend_cpp_with_app(args):
    """Extend CMAKE_PREFIX_PATH with AMENT_PREFIX_PATH."""
    warnings.warn(
        'colcon_ros.task.extend_cpp_with_app() will be removed in the future',
        DeprecationWarning, stacklevel=2)
    ament_prefix_path = os.environ.get('AMENT_PREFIX_PATH')
    if ament_prefix_path:
        ament_prefix_path = ament_prefix_path.replace(
            os.pathsep, ';')
        if args.cmake_args is None:
            args.cmake_args = []
        # check if the CMAKE_PREFIX_PATH is explicitly set
        prefix = '-DCMAKE_PREFIX_PATH='
        for i, value in reversed(list(enumerate(args.cmake_args))):
            if not value.startswith(prefix):
                continue
            # extend the last existing entry
            existing = value[len(prefix):]
            if existing:
                existing = ';' + existing
            args.cmake_args[i] = \
                '-DCMAKE_PREFIX_PATH={ament_prefix_path}{existing}' \
                .format_map(locals())
            break
        else:
            # otherwise extend the environment variable
            existing = os.environ.get('CMAKE_PREFIX_PATH', '')
            if existing:
                existing = ';' + existing.replace(
                    os.pathsep, ';')
            args.cmake_args.append(
                '-DCMAKE_PREFIX_PATH={ament_prefix_path}{existing}'
                .format_map(locals()))



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.prefix_path import PrefixPathExtensionPoint

logger = colcon_logger.getChild(__name__)

_get_ament_prefix_path_warnings = set()


class AmentPrefixPath(PrefixPathExtensionPoint):
    """Prefix path defined in the `AMENT_PREFIX_PATH` environment variable."""

    # the priority needs to be lower than the colcon prefix path extension
    PRIORITY = 90

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PrefixPathExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def extend_prefix_path(self, paths):  # noqa: D102
        global _get_ament_prefix_path_warnings
        ament_prefix_path = os.environ.get('AMENT_PREFIX_PATH', '')
        for path in ament_prefix_path.split(os.pathsep):
            if not path:
                continue
            if not os.path.exists(path):
                if path not in _get_ament_prefix_path_warnings:
                    logger.warning(
                        "The path '{path}' in the environment variable "
                        "AMENT_PREFIX_PATH doesn't exist"
                        .format_map(locals()))
                _get_ament_prefix_path_warnings.add(path)
                continue

            for filename in os.listdir(path):
                if filename.startswith('local_setup.'):
                    break
            else:
                parent_path = os.path.dirname(path)
                marker_file = os.path.join(
                    parent_path, '.colcon_install_layout')
                if not os.path.exists(marker_file):
                    if path not in _get_ament_prefix_path_warnings:
                        logger.warning(
                            "The path '{path}' in the environment variable "
                            "AMENT_PREFIX_PATH doesn't contain any "
                            "'local_setup.*' files.".format_map(locals()))
                        _get_ament_prefix_path_warnings.add(path)
                    continue
                with open(marker_file, 'r') as h:
                    install_layout = h.read().rstrip()
                if install_layout != 'isolated':
                    if path not in _get_ament_prefix_path_warnings:
                        logger.warning(
                            "The path '{path}' in the environment variable "
                            "AMENT_PREFIX_PATH doesn't use the expected "
                            "install layout 'isolated'.".format_map(locals()))
                        _get_ament_prefix_path_warnings.add(path)
                    continue
                path = parent_path

            paths.append(path)



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path

from colcon_cmake.task.cmake.test import CmakeTestTask
from colcon_core.environment import create_environment_scripts_only
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import create_environment_hook
from colcon_core.task import TaskExtensionPoint
from colcon_ros.task.catkin import create_pythonpath_environment_hook

logger = colcon_logger.getChild(__name__)


class CatkinTestTask(TaskExtensionPoint):
    """Test ROS packages with the build type 'catkin'."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    async def test(self):  # noqa: D102
        args = self.context.args
        logger.info(
            "Testing ROS package in '{args.path}' with build type 'catkin'"
            .format_map(locals()))

        # for catkin packages it is expected that the devel space
        # and not the install space is in the environment for testing
        self.context.dependencies[self.context.pkg.name] = \
            os.path.join(self.context.args.build_base, 'devel')

        # additional hooks
        additional_hooks = create_environment_hook(
            'ros_package_path', Path(args.build_base) / 'devel',
            self.context.pkg.name,
            'ROS_PACKAGE_PATH', args.path, mode='prepend')
        additional_hooks += create_pythonpath_environment_hook(
            args.build_base, Path(args.build_base) / 'devel',
            self.context.pkg.name)

        # generate the necessary setup files for the devel space
        create_environment_scripts_only(
            Path(args.build_base) / 'devel', self.context.pkg,
            additional_hooks=additional_hooks)

        # reuse CMake test task
        extension = CmakeTestTask()
        extension.set_context(context=self.context)
        return await extension.test()



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.prefix_path import PrefixPathExtensionPoint

logger = colcon_logger.getChild(__name__)

_get_cmake_prefix_path_warnings = set()


class CmakePrefixPath(PrefixPathExtensionPoint):
    """
    Prefix path defined in the `CMAKE_PREFIX_PATH` environment variable.

    The path must contain a `.catkin` file to be considered.
    """

    # the priority needs to be lower than the ament prefix path extension
    PRIORITY = 80

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PrefixPathExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def extend_prefix_path(self, paths):  # noqa: D102
        global _get_cmake_prefix_path_warnings
        cmake_prefix_path = os.environ.get('CMAKE_PREFIX_PATH', '')
        for path in cmake_prefix_path.split(os.pathsep):
            if not path:
                continue
            if not os.path.exists(path):
                if path not in _get_cmake_prefix_path_warnings:
                    logger.warning(
                        "The path '{path}' in the environment variable "
                        "CMAKE_PREFIX_PATH doesn't exist"
                        .format_map(locals()))
                    _get_cmake_prefix_path_warnings.add(path)
                continue
            if not os.path.exists(os.path.join(path, '.catkin')):
                continue
            for filename in os.listdir(path):
                if filename.startswith('local_setup.'):
                    break
            else:
                parent_path = os.path.dirname(path)
                marker_file = os.path.join(
                    parent_path, '.colcon_install_layout')
                if not os.path.exists(marker_file):
                    if path not in _get_cmake_prefix_path_warnings:
                        logger.warning(
                            "The path '{path}' in the environment variable "
                            'CMAKE_PREFIX_PATH seems to be a catkin workspace '
                            "but it doesn't contain any 'local_setup.*' files."
                            ' Maybe the catkin version is not up-to-date?'
                            .format_map(locals()))
                        _get_cmake_prefix_path_warnings.add(path)
                    continue
                with open(marker_file, 'r') as h:
                    install_layout = h.read().rstrip()
                if install_layout != 'isolated':
                    if path not in _get_cmake_prefix_path_warnings:
                        logger.warning(
                            "The path '{path}' in the environment variable "
                            "CMAKE_PREFIX_PATH doesn't use the expected "
                            "install layout 'isolated'.".format_map(locals()))
                        _get_cmake_prefix_path_warnings.add(path)
                    continue
                path = parent_path

            paths.append(path)



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import os
from pathlib import Path

from colcon_cmake.task.cmake.build import CmakeBuildTask
from colcon_core.environment import create_environment_scripts
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import create_environment_hook
from colcon_core.shell import get_shell_extensions
from colcon_core.task import TaskExtensionPoint
from colcon_ros.task.catkin import create_pythonpath_environment_hook
from colcon_ros.task.cmake import create_pkg_config_path_environment_hooks

logger = colcon_logger.getChild(__name__)


class CatkinBuildTask(TaskExtensionPoint):
    """Build ROS packages with the build type 'catkin'."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--catkin-cmake-args',
            nargs='*', metavar='*', type=str.lstrip,
            help="Pass arguments to 'catkin' packages. "
            'Arguments matching other options must be prefixed by a space')
        parser.add_argument(
            '--catkin-skip-building-tests',
            action='store_true',
            help="By default the 'tests' target of 'catkin' packages is "
            "invoked. If running 'colcon test' later isn't intended this can "
            'be skipped')

    async def build(self):  # noqa: D102
        args = self.context.args
        logger.info(
            "Building ROS package in '{args.path}' with build type 'catkin'"
            .format_map(locals()))

        # reuse CMake build task with additional logic
        extension = CmakeBuildTask()
        extension.set_context(context=self.context)

        # additional arguments
        if args.cmake_args is None:
            args.cmake_args = []
        args.cmake_args += ['-DCATKIN_INSTALL_INTO_PREFIX_ROOT=0']
        if args.test_result_base:
            # catkin appends the project name itself
            args.cmake_args.append(
                '-DCATKIN_TEST_RESULTS_DIR=' +
                os.path.dirname(args.test_result_base))
        if args.catkin_cmake_args:
            args.cmake_args += args.catkin_cmake_args

        # invoke the build
        additional_targets = []
        # if no specific target is specified consider building the 'tests'
        # target and continue if such a target doesn't exist
        if args.cmake_target is None:
            if not args.catkin_skip_building_tests:
                additional_targets.append('tests')
                args.cmake_target_skip_unavailable = True
        rc = await extension.build(
            skip_hook_creation=True, additional_targets=additional_targets)

        # for catkin packages add additional hooks after the package has
        # been built and installed depending on the installed files
        additional_hooks = create_environment_hook(
            'ros_package_path', Path(args.install_base), self.context.pkg.name,
            'ROS_PACKAGE_PATH', 'share', mode='prepend')
        additional_hooks += create_pythonpath_environment_hook(
            args.build_base, Path(args.install_base), self.context.pkg.name)
        additional_hooks += create_pkg_config_path_environment_hooks(
            Path(args.install_base), self.context.pkg.name)

        # register hooks created via catkin_add_env_hooks
        shell_extensions = get_shell_extensions()
        file_extensions = OrderedDict()
        for shell_extensions_same_prio in shell_extensions.values():
            for shell_extension in shell_extensions_same_prio.values():
                for file_extension in shell_extension.get_file_extensions():
                    file_extensions[file_extension] = shell_extension
        custom_hooks_path = Path(args.install_base) / \
            'share' / self.context.pkg.name / 'catkin_env_hook'
        for file_extension, shell_extension in file_extensions.items():
            file_extension_hooks = sorted(custom_hooks_path.glob(
                '*.{file_extension}'.format_map(locals())))
            if file_extension_hooks:
                try:
                    # try to set CATKIN_ENV_HOOK_WORKSPACE explicitly before
                    # sourcing these hooks
                    additional_hooks.append(
                        shell_extension.create_hook_set_value(
                            'catkin_env_hook_workspace',
                            Path(args.install_base), self.context.pkg.name,
                            'CATKIN_ENV_HOOK_WORKSPACE',
                            ''))
                except NotImplementedError:
                    # since not all shell extensions might implement this
                    pass
                additional_hooks += file_extension_hooks

        create_environment_scripts(
            self.context.pkg, args, additional_hooks=additional_hooks)

        # ensure that the install base has the marker file
        # identifying it as a catkin workspace
        marker = Path(args.install_base) / '.catkin'
        marker.touch(exist_ok=True)

        return rc



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0


from colcon_cmake.task.cmake.test import CmakeTestTask
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.task import TaskExtensionPoint

logger = colcon_logger.getChild(__name__)


class AmentCmakeTestTask(TaskExtensionPoint):
    """Test ROS packages with the build type 'ament_cmake'."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    async def test(self):  # noqa: D102
        args = self.context.args
        logger.info(
            "Testing ROS package in '{args.path}' with build type "
            "'ament_cmake'".format_map(locals()))

        # reuse CMake test task
        extension = CmakeTestTask()
        extension.set_context(context=self.context)
        return await extension.test()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_cmake.task.cmake.build import CmakeBuildTask
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import get_shell_extensions
from colcon_core.task import TaskExtensionPoint
from colcon_ros.task import add_app_to_cpp

logger = colcon_logger.getChild(__name__)


class AmentCmakeBuildTask(TaskExtensionPoint):
    """Build ROS packages with the build type 'ament_cmake'."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--ament-cmake-args',
            nargs='*', metavar='*', type=str.lstrip,
            help="Pass arguments to 'ament_cmake' packages. "
            'Arguments matching other options must be prefixed by a space')

    async def build(self):  # noqa: D102
        args = self.context.args
        logger.info(
            "Building ROS package in '{args.path}' with build type "
            "'ament_cmake'".format_map(locals()))

        # reuse CMake build task with additional logic
        extension = CmakeBuildTask()
        extension.set_context(context=self.context)

        # add a hook for each available shell
        additional_hooks = []
        shell_extensions = get_shell_extensions()
        file_extensions = []
        for shell_extensions_same_prio in shell_extensions.values():
            for shell_extension in shell_extensions_same_prio.values():
                file_extensions += shell_extension.get_file_extensions()
        for file_extension in sorted(file_extensions):
            additional_hooks.append(
                'share/{self.context.pkg.name}/local_setup.{file_extension}'
                .format_map(locals()))

        # additional arguments
        if args.test_result_base:
            if args.cmake_args is None:
                args.cmake_args = []
            # ament_cmake appends the project name itself
            args.cmake_args.append(
                '-DAMENT_TEST_RESULTS_DIR=' +
                os.path.dirname(args.test_result_base))
        if args.symlink_install:
            if args.cmake_args is None:
                args.cmake_args = []
            args.cmake_args.append('-DAMENT_CMAKE_SYMLINK_INSTALL=1')
        if args.ament_cmake_args:
            if args.cmake_args is None:
                args.cmake_args = []
            args.cmake_args += args.ament_cmake_args

        return await extension.build(
            environment_callback=add_app_to_cpp,
            additional_hooks=additional_hooks)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from unittest.mock import Mock

from colcon_ros.package_identification.ros import _create_metadata


def test_version_contraint_eq():
    mock = Mock()
    mock.version_gt = None
    mock.version_lt = None
    mock.version_lte = None
    mock.version_gte = None

    mock.version_eq = '1.1.1'

    metadata = _create_metadata(mock)
    assert len(metadata.keys()) == 1
    assert metadata['version_eq'] == '1.1.1'


def test_version_contraint_gt():
    mock = Mock()
    mock.version_eq = None
    mock.version_lt = None
    mock.version_lte = None
    mock.version_gte = None

    mock.version_gt = '1.1.2'

    metadata = _create_metadata(mock)
    assert len(metadata.keys()) == 1
    assert metadata['version_gt'] == '1.1.2'


def test_version_contraint_lt():
    mock = Mock()
    mock.version_eq = None
    mock.version_gt = None
    mock.version_lte = None
    mock.version_gte = None

    mock.version_lt = '1.1.5'

    metadata = _create_metadata(mock)
    assert len(metadata.keys()) == 1
    assert metadata['version_lt'] == '1.1.5'


def test_version_contraint_lte():
    mock = Mock()
    mock.version_eq = None
    mock.version_gt = None
    mock.version_lt = None
    mock.version_gte = None

    mock.version_lte = '10.1.3'

    metadata = _create_metadata(mock)
    assert len(metadata.keys()) == 1
    assert metadata['version_lte'] == '10.1.3'


def test_version_contraint_gte():
    mock = Mock()
    mock.version_eq = None
    mock.version_gt = None
    mock.version_lt = None
    mock.version_lte = None

    mock.version_gte = '10.45.5'

    metadata = _create_metadata(mock)
    assert len(metadata.keys()) == 1
    assert metadata['version_gte'] == '10.45.5'


def test_version_constraint_none():
    mock = Mock()
    mock.version_eq = None
    mock.version_gt = None
    mock.version_lt = None
    mock.version_lte = None
    mock.version_gte = None
    metadata = _create_metadata(mock)
    assert len(metadata.keys()) == 0



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import sys

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import create_environment_hook
from colcon_core.shell import get_command_environment
from colcon_core.task import create_file
from colcon_core.task import install
from colcon_core.task import TaskExtensionPoint
from colcon_core.task.python import get_data_files_mapping
from colcon_core.task.python import get_setup_data
from colcon_core.task.python.build import PythonBuildTask

logger = colcon_logger.getChild(__name__)


class AmentPythonBuildTask(TaskExtensionPoint):
    """Build ROS packages with the build type 'ament_python'."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    async def build(self):  # noqa: D102
        args = self.context.args
        logger.info(
            "Building ROS package in '{args.path}' with build type "
            "'ament_python'".format_map(locals()))

        # reuse Python build task with additional logic
        extension = PythonBuildTask()
        extension.set_context(context=self.context)

        # additional hooks
        additional_hooks = create_environment_hook(
            'ament_prefix_path', Path(args.install_base),
            self.context.pkg.name, 'AMENT_PREFIX_PATH', '', mode='prepend')

        # get options from the Python manifest
        try:
            env = await get_command_environment(
                'setup_py', args.build_base, self.context.dependencies)
        except RuntimeError as e:
            logger.error(str(e))
            return 1
        setup_py_data = get_setup_data(self.context.pkg, env)

        # check if the package index and manifest are being installed
        data_files = get_data_files_mapping(
            setup_py_data.get('data_files', []) or [])
        installs_package_index = False
        installs_package_manifest = False

        # check if package index and manifest are being installed
        for source, destination in data_files.items():
            if sys.platform == 'win32':
                destination = Path(destination).as_posix()
            # work around data files incorrectly defined as not relative
            if os.path.isabs(source):
                source = os.path.relpath(source, args.path)
            if (
                destination == 'share/ament_index/resource_index/packages/' +
                self.context.pkg.name
            ):
                installs_package_index = True
            elif (
                source == 'package.xml' and
                destination == 'share/{self.context.pkg.name}/package.xml'
                .format_map(locals())
            ):
                installs_package_manifest = True

        # warn about missing explicit installation
        # for now implicitly install the marker and the manifest
        if not installs_package_index:
            # TODO remove magic helper in the future
            logger.warn(
                "Package '{self.context.pkg.name}' doesn't explicitly install "
                'a marker in the package index (colcon-ros currently does it '
                'implicitly but that fallback will be removed in the future)'
                .format_map(locals()))
            # create package marker in ament resource index
            create_file(
                args,
                'share/ament_index/resource_index/packages/'
                '{self.context.pkg.name}'.format_map(locals()))
        if not installs_package_manifest:
            # TODO remove magic helper in the future
            logger.warn(
                "Package '{self.context.pkg.name}' doesn't explicitly install "
                "the 'package.xml' file (colcon-ros currently does it "
                'implicitly but that fallback will be removed in the future)'
                .format_map(locals()))
            # copy / symlink package manifest
            install(
                args, 'package.xml',
                'share/{self.context.pkg.name}/package.xml'
                .format_map(locals()))

        return await extension.build(additional_hooks=additional_hooks)



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import platform
import subprocess

from colcon_core.shell import create_environment_hook


def create_pkg_config_path_environment_hooks(basepath, pkg_name):
    """
    Create hook scripts to prepend to the PKG_CONFIG_PATH.

    :param Path basepath: The path of the prefix
    :param str pkg_name: The package name
    :returns: The relative paths to the created hook scripts
    :rtype: list
    """
    hooks = create_environment_hook(
        'pkg_config_path', basepath, pkg_name, 'PKG_CONFIG_PATH',
        os.path.join('lib', 'pkgconfig'), mode='prepend')
    if platform.system() == 'Linux':
        multiarch = get_multiarch()
        if multiarch:
            hooks += create_environment_hook(
                'pkg_config_path_multiarch', basepath, pkg_name,
                'PKG_CONFIG_PATH', os.path.join('lib', multiarch, 'pkgconfig'),
                mode='prepend')
    return hooks


_multiarch = None


def get_multiarch():
    """
    Get the multiarch name using normalized GNU triplets.

    :returns: The multiarch name or None
    :rtype: str
    """
    global _multiarch
    if _multiarch is None:
        try:
            output = subprocess.check_output(['gcc', '-print-multiarch'])
        except subprocess.CalledProcessError:
            pass
        else:
            _multiarch = output.decode().rstrip()
        if _multiarch is None:
            try:
                output = subprocess.check_output(
                    ['dpkg-architecture', '-qDEB_HOST_MULTIARCH'])
            except subprocess.CalledProcessError:
                pass
            else:
                _multiarch = output.decode().rstrip()
    return _multiarch



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path

from colcon_cmake.task.cmake.build import CmakeBuildTask as CmakeBuildTask_
from colcon_core.environment import create_environment_scripts
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.task import TaskExtensionPoint
from colcon_ros.task import add_app_to_cpp
from colcon_ros.task.cmake import create_pkg_config_path_environment_hooks

logger = colcon_logger.getChild(__name__)


class CmakeBuildTask(TaskExtensionPoint):
    """Build ROS packages with the build type 'cmake'."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    async def build(self):  # noqa: D102
        args = self.context.args
        logger.info(
            "Building ROS package in '{args.path}' with build type 'cmake'"
            .format_map(locals()))

        # reuse CMake build task with additional logic
        extension = CmakeBuildTask_()
        extension.set_context(context=self.context)

        rc = await extension.build(
            skip_hook_creation=True, environment_callback=add_app_to_cpp)

        additional_hooks = create_pkg_config_path_environment_hooks(
            Path(args.install_base), self.context.pkg.name)

        create_environment_scripts(
            self.context.pkg, args, additional_hooks=additional_hooks)

        return rc



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_cmake.task.cmake import get_variable_from_cmake_cache
from colcon_core.logging import colcon_logger
from colcon_core.shell import create_environment_hook

logger = colcon_logger.getChild(__name__)


def create_pythonpath_environment_hook(build_base, install_base, pkg_name):
    """
    Create a hook script for each primary shell to prepend to the PYTHONPATH.

    :param str build_base: The path of the build directory
    :param Path install_base: The path of the install prefix
    :param str pkg_name: The package name
    :returns: The relative paths to the created hook scripts
    :rtype: list
    """
    hooks = []
    # catkin packages might use `--install-layout deb`
    # therefore getting the used install directory from the CMake cache
    # since it might not match distutils.sysconfig.get_python_lib()
    rel_python_path = get_variable_from_cmake_cache(
        build_base, 'PYTHON_INSTALL_DIR')
    # prepend Python specific path to PYTHONPATH if it exists
    if rel_python_path:
        abs_python_path = install_base / rel_python_path
        logger.log(1, "checking '%s'" % abs_python_path)
        if abs_python_path.exists():
            hooks += create_environment_hook(
                'catkin_pythonpath', install_base, pkg_name,
                'PYTHONPATH', rel_python_path, mode='prepend')
    return hooks



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from setuptools import setup

cmdclass = {}
try:
    from stdeb.command.sdist_dsc import sdist_dsc
except ImportError:
    pass
else:
    class CustomSdistDebCommand(sdist_dsc):
        """Weird approach to apply the Debian patches during packaging."""

        def run(self):  # noqa: D102
            from stdeb.command import sdist_dsc
            build_dsc = sdist_dsc.build_dsc

            def custom_build_dsc(*args, **kwargs):
                nonlocal build_dsc
                debinfo = self.get_debinfo()
                repackaged_dirname = \
                    debinfo.source + '-' + debinfo.upstream_version
                dst_directory = os.path.join(
                    self.dist_dir, repackaged_dirname, 'debian', 'patches')
                os.makedirs(dst_directory, exist_ok=True)
                # read patch
                with open('debian/patches/setup.cfg.patch', 'r') as h:
                    lines = h.read().splitlines()
                print(
                    "writing customized patch '%s'" %
                    os.path.join(dst_directory, 'setup.cfg.patch'))
                # write patch with modified path
                with open(
                    os.path.join(dst_directory, 'setup.cfg.patch'), 'w'
                ) as h:
                    for line in lines:
                        if line.startswith('--- ') or line.startswith('+++ '):
                            line = \
                                line[0:4] + repackaged_dirname + '/' + line[4:]
                        h.write(line + '\n')
                with open(os.path.join(dst_directory, 'series'), 'w') as h:
                    h.write('setup.cfg.patch\n')
                return build_dsc(*args, **kwargs)

            sdist_dsc.build_dsc = custom_build_dsc
            super().run()
    cmdclass['sdist_dsc'] = CustomSdistDebCommand

setup(cmdclass=cmdclass)



# Copyright 2016-2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path

import pytest
from scspell import Report
from scspell import SCSPELL_BUILTIN_DICT
from scspell import spell_check


spell_check_words_path = Path(__file__).parent / 'spell_check.words'


@pytest.fixture(scope='module')
def known_words():
    global spell_check_words_path
    return spell_check_words_path.read_text().splitlines()


def test_spell_check(known_words):
    source_filenames = [Path(__file__).parents[1] / 'setup.py'] + \
        list((Path(__file__).parents[1] / 'colcon_ros').glob('**/*.py')) + \
        list((Path(__file__).parents[1] / 'test').glob('**/*.py'))

    for source_filename in sorted(source_filenames):
        print('Spell checking:', source_filename)

    # check all files
    report = Report(known_words)
    spell_check(
        [str(p) for p in source_filenames], base_dicts=[SCSPELL_BUILTIN_DICT],
        report_only=report, additional_extensions=[('', 'Python')])

    unknown_word_count = len(report.unknown_words)
    assert unknown_word_count == 0, \
        'Found {unknown_word_count} unknown words: '.format_map(locals()) + \
        ', '.join(sorted(report.unknown_words))

    unused_known_words = set(known_words) - report.found_known_words
    unused_known_word_count = len(unused_known_words)
    assert unused_known_word_count == 0, \
        '{unused_known_word_count} words in the word list are not used: ' \
        .format_map(locals()) + ', '.join(sorted(unused_known_words))


def test_spell_check_word_list_order(known_words):
    assert known_words == sorted(known_words), \
        'The word list should be ordered alphabetically'


def test_spell_check_word_list_duplicates(known_words):
    assert len(known_words) == len(set(known_words)), \
        'The word list should not contain duplicates'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_identification import IgnoreLocationException
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version


class IgnorePackageIdentification(PackageIdentificationExtensionPoint):
    """
    Ignore paths containing files belonging to `catkin` or `ament_tools`.

    Workspaces built with `catkin` or `ament_tools` don't contain any IGNORE
    marker files in the install space.
    To avoid packages being found in these directories by the discovery
    extensions using default values they is being ignored by this extension.
    """

    # the priority needs to be higher than all other extensions
    PRIORITY = 1000

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def identify(self, desc):  # noqa: D102
        filenames = (
            '_order_isolated_packages.py',  # ament_tools
            '_setup_util.py',  # catkin
        )
        for filename in filenames:
            ignore = desc.path / filename
            if ignore.exists():
                raise IgnoreLocationException()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

# try import since this package doesn't depend on colcon-argcomplete
# and needs to fallback gracefully if it is not available
try:
    from colcon_argcomplete.argcomplete_completer \
        import ArgcompleteCompleterExtensionPoint
except ImportError:
    class ArgcompleteCompleterExtensionPoint:  # noqa: D101
        pass
from colcon_cmake.argcomplete_completer.cmake_args \
    import get_cmake_args_completer_choices
from colcon_core.plugin_system import satisfies_version


class CatkinCmakeArgcompleteCompleter(ArgcompleteCompleterExtensionPoint):
    """Completion of catkin CMake arguments."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            ArgcompleteCompleterExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def get_completer(self, parser, *args, **kwargs):  # noqa: D102
        # only contribute a complete for CMake arguments
        if '--catkin-cmake-args' not in args:
            return None

        # check that the argcomplete version has the necessary completer
        try:
            from argcomplete.completers import ChoicesCompleter
        except ImportError:
            return None

        # since only one extension can contribute a completer
        # include the CMake choices in addition to the catkin ones
        return ChoicesCompleter(
            get_cmake_args_completer_choices() +
            get_catkin_cmake_args_completer_choices())


def get_catkin_cmake_args_completer_choices():
    """Get completer choices for CMake arguments for catkin packages."""
    # HACK the quote and equal characters are currently a problem
    # see https://github.com/kislyuk/argcomplete/issues/94
    return \
        [
            ' -DCATKIN_ENABLE_TESTING={flag}'.format_map(locals())
            for flag in ('OFF', 'ON')] + \
        [
            ' -DCATKIN_SKIP_TESTING={flag}'.format_map(locals())
            for flag in ('OFF', 'ON')] + \
        [
            ' -DSETUPTOOLS_DEB_LAYOUT={flag}'.format_map(locals())
            for flag in ('OFF', 'ON')]
